#include <engine/3d/bezier.h>
#include <LIMITS.H>

/*
 * set_key_buf:
 *	Initializes the caller's local key buffer.
 */
static
set_key_buf(key, p_bezier)
VECTOR *key; /* [MAX_BEZIER_KEY] */
struct Bezier3D *p_bezier;
{
	int i;
	for(i = 0; i < p_bezier->nkeys; i++)
	{
		key[i] = p_bezier->key[i];
		key[i].vx >>= 12;
		key[i].vy >>= 12;
		key[i].vz >>= 12;
	}
}

/*
 * track:
 *	Lerp 'til the cows come home :)
 */
static
track(key, t, n)
VECTOR *key; /* [MAX_BEZIER_KEY] */
int t, n;
{
	int i;

	for(i = 0; i < n-1; i++)
	{
		key[i].vx = lerpi(key[i].vx, key[i+1].vx, t, 16)>>16;
		key[i].vy = lerpi(key[i].vy, key[i+1].vy, t, 16)>>16;
		key[i].vz = lerpi(key[i].vz, key[i+1].vz, t, 16)>>16;
	}

	if(--n < 2)
	{
		return;
	}
	else
	{
		track(key, t, n);
	}
}

/*
 * bezier3d_track:
 *	Calculates the 3D coords at a specific
 *	point (t) along the curve. t has 16 fractional
 *	bits. Effectively acts as a wrapper for track()
 */
bezier3d_track(p_bezier)
struct Bezier3D *p_bezier;
{
	unsigned	i, j;
	VECTOR		key[MAX_BEZIER_KEY];

	if(!p_bezier->nkeys)
	{
		return;
	}

	set_key_buf(key, p_bezier);
	track(key, p_bezier->cursor.t, p_bezier->nkeys);

	setVector(&p_bezier->cursor.coords, key[0].vx, key[0].vy, key[0].vz);
}

/*
 * bezier3d_init:
 *	Initializes a Bezier3D struct from a BEZ file.
 */
bezier3d_init(p_bezier, p_data, radius)
struct Bezier3D *p_bezier;
void *p_data;
int radius;
{
	int		i;
	int		*p_ptr = (int *)p_data;

	VECTOR		key[MAX_BEZIER_KEY];

	p_bezier->nkeys = *p_ptr;
	p_ptr++;

	if(p_bezier->nkeys < 2 || (p_bezier->nkeys > MAX_BEZIER_KEY))
	{
		printf("bezier3d_init() error: number of keys out of range! (%d)\n", p_bezier->nkeys);
		while(1);
	}

	for(i = 0; i < p_bezier->nkeys; i++, p_ptr+=3)
	{
		setVector(p_bezier->key+i, *p_ptr, *(p_ptr+1), *(p_ptr+2));

		SetTile(p_bezier->keyprim+i);
		setWH(p_bezier->keyprim+i, 4, 4);
		setRGB0(p_bezier->keyprim+i, 255, 0, 0);
	}

	for(i = 0; i < MAX_BEZIER_PRIM; i++)
	{
		SetTile(p_bezier->tile+i);
		setWH(p_bezier->tile+i, 4, 4);
		setRGB0(p_bezier->tile+i, 255, 255, 255);

		set_key_buf(key, p_bezier);

		track(key, (i<<18)/(MAX_BEZIER_PRIM<<2), p_bezier->nkeys);
		setVector(p_bezier->tilepos+i, key[0].vx, key[0].vy, key[0].vz);
	}

	SetTile(&p_bezier->cursor.tile);
	setWH(&p_bezier->cursor.tile, 4, 4);
	setRGB0(&p_bezier->cursor.tile, 255, 0, 255);
	p_bezier->cursor.t = 0;

	p_bezier->col.prev = 0;
	p_bezier->col.cur = 0;
}

/*
 * bezier3d_collide_d:
 *	3D bezier collision detection - checks whether or not
 *	a point is located inside a curve.
 */
bezier3d_collide_d(p_bezier, radius, p_pos)
struct Bezier3D *p_bezier;
unsigned int radius; /* FIXME: just use the radius from the Bezier3D struct?!?! */
SVECTOR *p_pos;
{
	int			i;
	int			primidx = 0;
	unsigned int		r, low = UINT_MAX;
	unsigned int		distx, disty, distz;

	VECTOR			key[MAX_BEZIER_KEY];

	p_bezier->col.prev = p_bezier->col.cur;

	for(i = 0; i < MAX_BEZIER_PRIM; i++)
	{
		distx = ABS(p_pos->vx + p_bezier->tilepos[i].vx);
		disty = ABS(p_pos->vy + p_bezier->tilepos[i].vy);
		distz = ABS(p_pos->vz + p_bezier->tilepos[i].vz);
		r = psqrt( (ppow(distx, 2)>>6) + (ppow(disty, 2)>>6) + (ppow(distz, 2)>>6) );

		if(r < low)
		{
			primidx = i;
			low = r;
		}
	}

	p_bezier->cursor.t = primidx*BEZIER_UNIT;
	setVector(&p_bezier->cursor.coords, p_bezier->tilepos[primidx].vx, p_bezier->tilepos[primidx].vy, p_bezier->tilepos[primidx].vz);

	if(low > radius)
	{
		p_bezier->col.cur = 0;
		return 0;
	}

	p_bezier->col.cur = 1;
	return 1;
}

/*
 * bezier3d_collide_r:
 *	3D bezier collision response - checks if the bezier
 *	at its current state is valid or not.
 */
bezier3d_collide_r(p_bezier)
struct Bezier3D *p_bezier;
{
	int diff0 = ABS(p_bezier->cursor.t - 0);
	int diff1 = ABS(p_bezier->cursor.t - USHRT_MAX);

	if(diff0 < (BEZIER_UNIT*2) || diff1 < (BEZIER_UNIT*2))
	{
		return 0;
	}

	if(p_bezier->col.cur != p_bezier->col.prev)
	{
		printf("bezier3d_collide_r: %d | %d\n", diff0, diff1);
		return 1;
	}
	return 0;
}

